#ifndef _LINUX_KERNEL_H_
#define _LINUX_KERNEL_H_

#include <sys/cdefs.h>
__FBSDID("src/sys/compat/linux/linux_kernel.h,v 1.0 2005/04/16 11:12:26 Li Wang Exp $");

#include <sys/param.h>
#include <sys/module.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/sysproto.h>
#include <sys/fcntl.h>
#include <sys/file.h>
#include <sys/limits.h>
#include <sys/malloc.h>
#include <sys/queue.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/syscallsubr.h>
#include <sys/uio.h>
#include <sys/syslog.h>
#include <sys/smp.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/sx.h>

#include <net/if.h>
#include <net/route.h>
#include <net/pfil.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>

#include <machine/in_cksum.h>

#include <compat/linux/linux_list.h>            /* Linux list implementation */

#define __u16		u_int16_t
#define __u32		u_int32_t
#define __u64		u_int64_t

#define u16			u_int16_t
#define u32			u_int32_t
#define u64			u_int64_t

#define __s32 int32_t
#define __s16 int16_t
#define __s8  int8_t

typedef volatile u_int32_t atomic_t;
typedef struct uma_zone kmem_cache_t;

/*
 * GFP bitmasks..
 */
/* Zone modifiers in GFP_ZONEMASK (see linux/mmzone.h - low four bits) */
#define __GFP_DMA	0x01
#define __GFP_HIGHMEM	0x02

/* Action modifiers - doesn't change the zoning */
#define __GFP_WAIT	0x10	/* Can wait and reschedule? */
#define __GFP_HIGH	0x20	/* Should access emergency pools? */
#define __GFP_IO	0x40	/* Can start low memory physical IO? */
#define __GFP_HIGHIO	0x80	/* Can start high mem physical IO? */
#define __GFP_FS	0x100	/* Can call down to low-level FS? */

#define GFP_NOHIGHIO	(__GFP_HIGH | __GFP_WAIT | __GFP_IO)
#define GFP_NOIO	(__GFP_HIGH | __GFP_WAIT)
#define GFP_NOFS	(__GFP_HIGH | __GFP_WAIT | __GFP_IO | __GFP_HIGHIO)
#define GFP_ATOMIC	(__GFP_HIGH)
#define GFP_USER	(             __GFP_WAIT | __GFP_IO | __GFP_HIGHIO | __GFP_FS)
#define GFP_HIGHUSER	(             __GFP_WAIT | __GFP_IO | __GFP_HIGHIO | __GFP_FS | __GFP_HIGHMEM)
#define GFP_KERNEL	(__GFP_HIGH | __GFP_WAIT | __GFP_IO | __GFP_HIGHIO | __GFP_FS)
#define GFP_NFS		(__GFP_HIGH | __GFP_WAIT | __GFP_IO | __GFP_HIGHIO | __GFP_FS)
#define GFP_KSWAPD	(             __GFP_WAIT | __GFP_IO | __GFP_HIGHIO | __GFP_FS)

/* Flag - indicates that the buffer will be suitable for DMA.  Ignored on some
   platforms, used as appropriate on others */

#define GFP_DMA		__GFP_DMA

/* flags to pass to kmem_cache_create().
 * The first 3 are only valid when the allocator as been build
 * SLAB_DEBUG_SUPPORT.
 */
#define	SLAB_DEBUG_FREE		0x00000100UL	/* Peform (expensive) checks on free */
#define	SLAB_DEBUG_INITIAL	0x00000200UL	/* Call constructor (as verifier) */
#define	SLAB_RED_ZONE		0x00000400UL	/* Red zone objs in a cache */
#define	SLAB_POISON		0x00000800UL	/* Poison objects */
#define	SLAB_NO_REAP		0x00001000UL	/* never reap from the cache */
#define	SLAB_HWCACHE_ALIGN	0x00002000UL	/* align objs on a h/w cache lines */
#define SLAB_CACHE_DMA		0x00004000UL	/* use GFP_DMA memory */
#define SLAB_MUST_HWCACHE_ALIGN	0x00008000UL	/* force alignment */

#define PROT_SOCK	1024

#define atomic_read(p)		(*(p))
#define atomic_set(p, v)	(*(p) = (v))
#define atomic_inc(p)		atomic_add_int(p, 1)
#define atomic_dec(p)	atomic_subtract_int(p, 1)
#define atomic_add(n, p)	atomic_add_int(p, n)
#define atomic_sub(n, p)	atomic_subtract_int(p, n)

#define NIPQUAD(addr) \
	((unsigned char *)&addr)[0], \
	((unsigned char *)&addr)[1], \
	((unsigned char *)&addr)[2], \
	((unsigned char *)&addr)[3]

static inline u_int atomic_dec_and_test(atomic_t *p)
{
	atomic_dec(p);
	return atomic_read(p)==0;
}
	
#define HZ	hz

/* 
  *  Use mtx_lock() to simulate down_interruptible()
  *  Fix me:How to return EINTR,turnstile_wait() return void
  */
static inline int down_interruptible(struct mtx * sem)
{
	mtx_lock(sem);
	return 0;
}

#define down(x)	down_interruptible(x)

static inline void up(struct mtx * sem)
{
	mtx_unlock(sem);
}

/*
  * Fix me:use sx_slock(),sx_xlock() to simulate read_lock()/write_lcok() 
  * is safe enough?the performance is whether better than mtx_lock()?
  * 
  */
#define read_lock(x)				sx_slock(x)
#define read_unlock(x)			sx_sunlock(x)

#define read_lock_bh(x)			sx_slock(x)
#define read_unlock_bh(x)     		sx_sunlock(x)

#define write_lock(x)				sx_xlock(x)
#define write_unlock(x)			sx_xunlock(x)

#define write_lock_bh(x)			sx_xlock(x)
#define write_unlock_bh(x)		sx_xunlock(x)

#define spin_lock_bh(x)			mtx_lock(x)
#define spin_unlock_bh(x)			mtx_unlock(x)

#define copy_from_user(dst,src,size)		copyin(src,dst,size)
#define copy_to_user(dst,src,size)			copyout(src,dst,size)

#define	memmove(T, F, L) 				bcopy((F), (T), (L))
#define	memcpy(dst,src,len)				bcopy((src),(dst),(len))

static __inline int after(__u32 seq1, __u32 seq2)
{
	return (__s32)(seq2-seq1) < 0;
}

#define 	net_msg_cost 	5*hz
#define	net_msg_burst	10*5*hz

int net_ratelimit(void);

static inline kmem_cache_t *
kmem_cache_create (char *name, size_t size, size_t offset,
	unsigned long flags, void (*ctor)(void*, kmem_cache_t *, unsigned long),
	void (*dtor)(void*, kmem_cache_t *, unsigned long))
{
	return uma_zcreate(name,size,NULL,NULL,NULL,NULL,UMA_ALIGN_PTR,0);
}

static inline void * kmem_cache_alloc (kmem_cache_t *cachep, int flags)
{
	if (flags&GFP_ATOMIC)
		return uma_zalloc(cachep,M_NOWAIT);
	else
		return uma_zalloc(cachep,M_WAITOK);
}

static inline void kmem_cache_free (kmem_cache_t *cachep, void *objp)
{
	return uma_zfree(cachep,objp);
}

static inline int kmem_cache_destroy (kmem_cache_t * cachep)
{
	uma_zdestroy(cachep);
	return 0;
}

enum
{
	RTN_UNSPEC,
	RTN_UNICAST,		/* Gateway or direct route	*/
	RTN_BROADCAST,		/* Accept locally as broadcast,
				   send as broadcast */
	RTN_MULTICAST,		/* Multicast route		*/
};

#define LOOPBACK(x)	(((x) & htonl(0xff000000)) == htonl(0x7f000000))
#define MULTICAST(x)	(((x) & htonl(0xf0000000)) == htonl(0xe0000000))
#define BADCLASS(x)	(((x) & htonl(0xf0000000)) == htonl(0xf0000000))
#define ZERONET(x)	(((x) & htonl(0xff000000)) == htonl(0x00000000))
#define LOCAL_MCAST(x)	(((x) & htonl(0xFFFFFF00)) == htonl(0xE0000000))

static inline unsigned inet_addr_type(u32 addr)
{
	unsigned ret = RTN_UNICAST;

	if (ZERONET(addr) || BADCLASS(addr))
		return RTN_BROADCAST;
	if (MULTICAST(addr))
		return RTN_MULTICAST;
	return ret;
}

#endif
